@@ -3,8 +3,8 @@
  * This file is part of Nothing (yet).
  *
  * $RCSfile: mod_pyapache.c,v $
- * $Revision: 2.21 $
- * $Date: 1996/11/11 12:54:00 $
+ * $Revision: 2.25 $
+ * $Date: 1997/01/22 13:31:48 $
  *
  * Created Fri May  3 15:12:20 1996.
  *
@@ -32,14 +32,23 @@
         it in several chunks for readability):
 
         PYTHONDIR= /usr/local/lib/python1.4/config
+        EXTRA_CFLAGS= -I$(PYTHONDIR)
         EXTRA_LIBS= $(PYTHONDIR)/config.o $(PYTHONDIR)/getpath.o \
                     $(PYTHONDIR)/libModules.a $(PYTHONDIR)/libPython.a \
                     $(PYTHONDIR)/libObjects.a $(PYTHONDIR)/libParser.a
         Module pyapache_module mod_pyapache.o
 
-        of course adapting PYTHONDIR value to your situation. Then you
-        should add `-I$(PYTHONDIR)' to the CFLAGS variable.
+        of course adapting PYTHONDIR value to your situation.
 
+	If your compiler does not already look into /usr/local/include
+	(where most probably resides the `python-1.4' include directory),
+	you should add `-I/usr/local/include' to EXTRA_CFLAGS.
+	
+	On Fri, 17 Jan 1997, Kent Polk <kent@eaenki.nde.swri.edu> said
+	that on Solaris 2.5.1 he had to explicitly add libdl.so and
+	libm.so to EXTRA_LIBS.
+	
+	
         Please note that you should execute a `make libainstall' in
         the Python source tree to get `getpath.c' and `config.c'
         installed (of course, this must be done *after* a successful
@@ -67,6 +76,7 @@
         directory, and so on):
 
         PYTHONDIR= /LocalDeveloper/WiP/Python-1.4
+        EXTRA_CFLAGS= -I$(PYTHONDIR)
         EXTRA_LIBS= $(PYTHONDIR)/Modules/config.o \
                     $(PYTHONDIR)/Modules/getpath.o \
                     $(PYTHONDIR)/Modules/libModules.a \
@@ -169,10 +179,12 @@
      the compiled scripts whenever you install their sources (see the
      `compileall.py' library module in the Python Distribution).
 
-   * This module requires Python 1.4 and Apache 1.1[.1], and will
-     require some work to adapt it to higher releases.
+   * This module requires Python 1.4 and Apache 1.2b4, and will
+     require some work to adapt it to higher releases. If you are still
+     using Apache 1.1[.1], then you need release 2.21 of this module.     
 
-   * Suspicious code is marked with a 'XXX' sign. */
+   * Suspicious code is marked with a 'XXX' sign: feel free to email me
+     your mumblings on it. */
 
 
 /* Apache includes */
@@ -197,7 +209,14 @@
 /* ...they miss this */
 extern long PyOS_GetLastModificationTime (char *path);
 
-/* Apache cries for autoconf!! */
+
+/* Apache cries for autoconf!!. Till then, you should check the
+   following manually. For example Solaris 2.5.1 needs this
+   define enabled */
+#if 0
+#define HAVE_ALLOCA_H
+#endif
+
 
 /* AIX requires this to be the first thing in the file.  */
 #ifdef __GNUC__
@@ -272,7 +291,7 @@ create_python_dir_config (pool *p, char *dummy)
 
 /* This function gets called at configuration time to handle `PythonPath'
    directives. */
-static char *
+static const char *
 set_python_path (cmd_parms *parms, python_dir_config *dc, char *arg)
 {
   dc->path = pstrdup (parms->pool, arg);
@@ -298,7 +317,7 @@ get_python_path (request_rec *r)
 
 /* This function gets called at configuration time when a PythonDebug
    directive is found. */
-static char *
+static const char *
 set_python_debug (cmd_parms *parms, python_dir_config *dc, char *arg)
 {
   if (!strcasecmp (arg, "off"))
@@ -313,7 +332,7 @@ set_python_debug (cmd_parms *parms, python_dir_config *dc, char *arg)
 
 /* This function gets called at configuration time when a
    PythonVerbose directive is found. */
-static char *
+static const char *
 set_python_verbose (cmd_parms *parms, python_dir_config *dc, char *arg)
 {
   if (!strcasecmp (arg, "off"))
@@ -328,7 +347,7 @@ set_python_verbose (cmd_parms *parms, python_dir_config *dc, char *arg)
 
 /* This function gets called at configuration time when a
    PythonPersistent directive is found. */
-static char *
+static const char *
 set_python_persistent (cmd_parms *parms, python_dir_config *dc, char *arg)
 {
   if (!strcasecmp (arg, "off"))
@@ -343,7 +362,7 @@ set_python_persistent (cmd_parms *parms, python_dir_config *dc, char *arg)
 
 /* This function gets called at configuration time when a
    PythonTrustedScripts directive is found. */
-static char *
+static const char *
 set_python_trusted_scripts (cmd_parms *parms, python_dir_config *dc, char *arg)
 {
   if (!PythonPersistent)
@@ -381,19 +400,20 @@ static PyObject *initialize_apache_module (request_rec *r);
 typedef struct python_child_stuff
 {
   request_rec *r;
+  pid_t pid;
   char *argv0;                  /* This is r->filename with the path stripped off */
-  FILE *script_in;              /* This is attached to the script's stdin */
+  FILE *script_in;              /* This gets attached to the script's stdin */
   FILE *script_out;             /* and this to its stdout, if not NPH. */
-  int args_size;                /* If > 0 this is an M_PUT or M_POST request */
+  FILE *script_err;             /* This gets attached to the script's stderr. */
   int nph;                      /* If != 0 this is an NPH script */
 } python_child_stuff;
 
-/* Validates the script, returns OK or the error code. */
+/* Validates the script, initializing the `argv0' and `nph' slots of
+   PCSP. Returns OK or the error code. */
 static int
 check (python_child_stuff *pcsp)
 {
   request_rec *r = pcsp->r;
-  int is_included = !strcmp (r->protocol, "INCLUDED");
 
   if ((pcsp->argv0 = strrchr (pcsp->r->filename, '/')) != NULL)
     pcsp->argv0++;
@@ -405,7 +425,13 @@ check (python_child_stuff *pcsp)
      "as is" to the client. */
   pcsp->nph = !(strncmp (pcsp->argv0, "nph-", 4));
   
-  if (pcsp->nph && is_included)
+  if (!(allow_options (r) & OPT_EXECCGI) && !is_scriptaliased (r))
+    {
+      log_reason("Option ExecCGI is off in this directory", r->filename, r);
+      return FORBIDDEN;
+    }
+  
+  if (pcsp->nph && !strcmp (r->protocol, "INCLUDED"))
     {
       log_reason("attempt to include NPH CGI script", r->filename, r);
       return FORBIDDEN;
@@ -428,22 +454,14 @@ check (python_child_stuff *pcsp)
       log_reason("file permissions deny server execution", r->filename, r);
       return FORBIDDEN;
     }
-  
-  if (r->method_number == M_POST || r->method_number == M_PUT)
-    {
-      const char *content_length = table_get (r->headers_in, "Content-length");
 
-      if (!content_length)
-        {
-          log_reason("POST or PUT without Content-length:", r->filename, r);
-          return BAD_REQUEST;
-        }
-      else
-        pcsp->args_size = atoi (content_length);
-    }
-  else
-    pcsp->args_size = -1;       /* means no args */
-  
+  {
+    int retval;
+
+    if ((retval = setup_client_block (r, REQUEST_CHUNKED_ERROR)))
+      return retval;
+  }
+    
   return OK;
 }
 
@@ -559,6 +577,7 @@ exec_the_script (python_child_stuff *pcsp)
       if (! n)
         {
           LOG_REASON ("Cannot parse the Python Script", pcsp->r->filename, pcsp->r);
+	  PyErr_Print();
           return SERVER_ERROR;
         }
 
@@ -568,6 +587,7 @@ exec_the_script (python_child_stuff *pcsp)
       if (! code)
         {
           LOG_REASON ("Cannot compile the Python Script", pcsp->r->filename, pcsp->r);
+	  PyErr_Print();
           return SERVER_ERROR;
         }
 
@@ -673,7 +693,7 @@ exec_the_script (python_child_stuff *pcsp)
 /* Initializes the Python Environment, the load path, the arguments list
    and its process environment.
 
-   If PythonTrustedScripts is On, then do most of the work just one. */
+   If PythonTrustedScripts is On, then do most of the work just once. */
 static int
 initialize_python_environment (python_child_stuff *pcsp)
 {
@@ -718,12 +738,14 @@ initialize_python_environment (python_child_stuff *pcsp)
     PySys_SetPath ((char *) ppath);
 
   /* Set the command line arguments list */
-  if ((!pcsp->r->args) || (!pcsp->r->args[0]) || (ind(pcsp->r->args,'=') >= 0))
-    PySys_SetArgv (1, &pcsp->r->filename);
+  if (!pcsp->r->args || !pcsp->r->args[0] || (ind (pcsp->r->args, '=') >= 0))
+    /* The aren't command line arguments: just set the first one to the
+       script's name */
+    PySys_SetArgv (1, &pcsp->argv0);
   else
     {
       int argc;
-      char ** argv = create_argv (pcsp->r->pool, pcsp->argv0, pcsp->r->args);
+      char ** argv = create_argv (pcsp->r, pcsp->argv0, pcsp->r->args, NULL);
 
       for (argc=0; argv[argc]; argc++)
         /* do nothing */;
@@ -770,7 +792,6 @@ initialize_python_environment (python_child_stuff *pcsp)
             PyDict_Clear (data);
           else
             {
-              /* XXX is this necessary ? */
               log_reason ("Unhandled environ type: %s",
                           environ_dict->ob_type->tp_name, pcsp->r);
               Py_DECREF (os_module);
@@ -869,7 +890,7 @@ copy_client_result (python_child_stuff *pcsp)
 {
   int ret;
 
-  if ((ret = scan_script_header (pcsp->r, pcsp->script_out)) == 0)
+  if ((ret = scan_script_header_err (pcsp->r, pcsp->script_out, NULL)) == OK)
     {
       char *location = table_get (pcsp->r->headers_out, "Location");
       
@@ -880,7 +901,9 @@ copy_client_result (python_child_stuff *pcsp)
           /* Soak up all the script output */
           hard_timeout ("read from script", pcsp->r);
           while (fgets (buffer, HUGE_STRING_LEN-1, pcsp->script_out) != NULL)
-            continue;
+            /* do nothing */;
+          while (fgets (buffer, HUGE_STRING_LEN-1, pcsp->script_out) != NULL)
+            /* do nothing */;
           kill_timeout (pcsp->r);
                       
           /* This redirect needs to be a GET no matter what the original
@@ -901,15 +924,22 @@ copy_client_result (python_child_stuff *pcsp)
         }
       else
         {
+          char buffer[HUGE_STRING_LEN];
+
           hard_timeout ("send script output", pcsp->r);
           
           send_http_header (pcsp->r);
           if (!pcsp->r->header_only)
             send_fd (pcsp->script_out, pcsp->r);
+
+          /* Soak up stderr */
+          while (fgets (buffer, HUGE_STRING_LEN-1, pcsp->script_err) != NULL)
+            /* do nothing */;
           
           kill_timeout (pcsp->r);
 
           pfclose (pcsp->r->connection->pool, pcsp->script_out);
+          pfclose (pcsp->r->connection->pool, pcsp->script_err);
 
           ret = OK;
         }
@@ -929,12 +959,13 @@ child_execute_python_script (python_child_stuff *pcsp)
      pass a pointer to script_out, since we will not be interested in it:
      the script's output is connected directly to the client. */
   
-  if (! spawn_child (pcsp->r->connection->pool,
-                     python_child,
-                     (void *) pcsp,
-                     pcsp->nph ? just_wait : kill_after_timeout,
-                     &pcsp->script_in,
-                     pcsp->nph ? NULL : &pcsp->script_out))
+  if (! (pcsp->pid = spawn_child_err (pcsp->r->connection->pool,
+                                      python_child,
+                                      (void *) pcsp,
+                                      pcsp->nph ? just_wait : kill_after_timeout,
+                                      &pcsp->script_in,
+                                      pcsp->nph ? NULL : &pcsp->script_out,
+                                      &pcsp->script_err)))
     {
       log_reason ("couldn't spawn child process", pcsp->r->filename, pcsp->r);
       return SERVER_ERROR;
@@ -942,27 +973,25 @@ child_execute_python_script (python_child_stuff *pcsp)
 
   /* If there are args coming from the client, forward them to the
      script' stdin */
-  if (pcsp->args_size > 0)      /* M_POST || M_PUT */
+  if (should_client_block (pcsp->r))
     {
       void (*handler)();
-      int remaining = pcsp->args_size;
+      int len_read;
       char argsbuffer[HUGE_STRING_LEN];
       
       hard_timeout ("copy script args", pcsp->r);
       handler = signal (SIGPIPE, SIG_IGN);
     
-      while ((remaining > 0))
+      while (((len_read =
+               get_client_block (pcsp->r, argsbuffer, HUGE_STRING_LEN)) > 0))
         {
-          int len_read, len_to_read = remaining;
-          
-          if (len_to_read > HUGE_STRING_LEN)
-            len_to_read = HUGE_STRING_LEN;
-          
-          len_read = read_client_block (pcsp->r, argsbuffer, len_to_read);
-          if (len_read == 0)
-            break;
-          fwrite (argsbuffer, 1, len_read, pcsp->script_in);
-          remaining -= len_read;
+          if (fwrite (argsbuffer, 1, len_read, pcsp->script_in) < (size_t) len_read)
+            {
+              /* silly script stopped reading, soak up remaining message */
+              while (get_client_block (pcsp->r, argsbuffer, HUGE_STRING_LEN) > 0)
+                /* do nothing */;
+              break;
+            }
         }
     
       fflush (pcsp->script_in);
@@ -976,7 +1005,10 @@ child_execute_python_script (python_child_stuff *pcsp)
   /* If the script is NPH our job finishes here; otherwise parse the
      incoming headers and copy the result to the client. */
   if (pcsp->nph)
-    return OK;
+    {
+      waitpid (pcsp->pid, (int *) 0, 0);
+      return OK;
+    }
   else
     return copy_client_result (pcsp);
 }
@@ -1128,7 +1160,7 @@ python_cgi_handler (request_rec *r)
 
   pcs.r = r;
   /* These other members of pcs get initialized by check():
-     pcs.argv0, pcs.args_size, pcs.nph */
+     pcs.argv0, pcs.nph */
 
   if ((ret = check (&pcs)) != OK)
     return ret;
